"""
Restae settings file
"""


class Settings(dict):
    """
    Settings class dict-like
    """
    def override(self, _settings):
        """
        Override basic settings
        """
        self.update(_settings)


DEFAULT_PAGE_SIZE = 20
DEFAULT_PAGINATION_CLASS = 'restae.pagination.CursorPagination'

RESTAE_SETTINGS = {
    'MIDDLEWARE_CLASSES': (
        ''
    ),
    'PAGINATION_CLASS': 'restae.pagination.CursorPagination',
    'PAGE_SIZE': DEFAULT_PAGE_SIZE,

    'CORS_ALLOW_METHODS': [
       'POST', 'GET', 'PUT', 'PATCH', 'DELETE'
    ],
    'CORS_ALLOW_HEADERS': [
        'Authorization',
        'Access-Control-Allow-Headers',
        'Origin',
        'Accept',
        'X-Requested-With',
        'Content-Type',
        'Access-Control-Request-Method',
        'Access-Control-Request-Headers'
    ],
    'CORS_ALLOW_ORIGIN': ['*']
}

settings = Settings()
settings.override(RESTAE_SETTINGS)



"""
Exceptions classes
"""


class NotFound(Exception):
    pass


class NotAuthorized(Exception):
    pass


class Forbidden(Exception):
    pass


class MissingParameter(Exception):
    pass


class BadRequest(Exception):
    pass


class MissingBody(Exception):
    pass


class ValidationError(Exception):
    pass


class SerializerError(Exception):
    pass


class DispatchError(Exception):
    pass


class InvalidPage(Exception):
    pass


class ImproperlyConfigured(Exception):
    pass



"""
Handlers modules
"""
import json
import re
import traceback

import webapp2
import logging

from webob import Response

from exceptions import NotFound, NotAuthorized, Forbidden, MissingParameter, BadRequest, MissingBody
from response import CorsResponse, JsonResponse
from helpers import get_object_from_urlsafe, get_middlewares, cached_property, get_key_from_urlsafe, \
    get_model_class_from_query, load_class
from restae.conf import DEFAULT_PAGINATION_CLASS
from restae.exceptions import DispatchError


class BaseHandler(webapp2.RequestHandler):
    """
    BaseHandler with some intelligence & automation
    """
    model = None
    lookup_field = None
    query_param = None

    def __init__(self, *args, **kwargs):
        super(BaseHandler, self).__init__(*args, **kwargs)
        self.route_args = {}
        self.env = {}
        self.query_params = {}
        self.user = None
        self.action = None

    @cached_property
    def middlewares(self):
        """
        Cached property that load all middlewares

        :return: :py:class:`list` of :py:class:`Middleware` instances
        """
        return get_middlewares()

    def dispatch(self):
        """
        Function called to make the dispatch job.
        Calls internally :py:meth:`restae.handlers.BaseHandler.apply_dispatch`

        :return: :py:class:`webob.Response`
        """
        self.action = self.request.method.lower()

        return self.apply_dispatch()

    def apply_dispatch(self):
        """
        Does the hole mechanism of dispatch.

        Loops over all middlewares before and after the dispatch method call


        .. code-block:: python

            for middleware in self.middlewares:
                middleware.process_request(self.request)

            self.route_args = self.get_route_args()
            response = self.do_dispatch()

            for middleware in self.middlewares:
                middleware.process_response(self.request, response)

        :return: :py:class:`webob.Response`
        """
        try:
            self.query_params = dict(self.request.GET)
            self.env = self.request.GET.env

            for middleware in self.middlewares:
                if middleware.activate_on_method(self.request.method.upper()):
                    middleware.process_request(self.request)

            self.route_args = self.get_route_args()
            response = self.do_dispatch()

            for middleware in self.middlewares:
                if middleware.activate_on_method(self.request.method.upper()):
                    middleware.process_response(self.request, response)

            return response
        except NotFound as nf:
            return JsonResponse(status=404, data=str(nf) or 'Not found')
        except NotAuthorized as na:
            return JsonResponse(status=403, data=str(na) or 'Not authorized')
        except Forbidden as fo:
            return JsonResponse(status=401, data=str(fo) or 'Forbidden')
        except MissingParameter as mp:
            return JsonResponse(status=400, data=str(mp) or 'Missing parameter')
        except BadRequest as br:
            return JsonResponse(status=400, data=str(br) or 'Bad request')
        except MissingBody as mb:
            return JsonResponse(status=400, data=str(mb) or 'Request is missing a body')
        except ValueError as ve:
            return JsonResponse(status=400, data=str(ve) or 'Value error')
        except DispatchError:
            return Response(status=404)
        except Exception as err:
            logging.error('%s: %s',
                          err.__class__.__name__,
                          str(err))
            logging.error(traceback.format_exc())
            return self.handle_exception(err, self.app.debug)

    def check_action_permissions(self):
        return True

    def do_dispatch(self):
        """
        Function that does the dispatch job

        :return: :py:class:`webob.Response`
        """
        if self.check_action_permissions() is False:
            raise NotAuthorized()

        return super(BaseHandler, self).dispatch()

    def get_body(self):
        """
        Tries to parse the request body as JSON

        :return: JSON serialized data as :py:class:`dict` instance.
        """
        _cached_body = getattr(self, '__cached_body', None)

        if _cached_body is not None:
            return _cached_body

        if self.request.body:
            try:
                json_body = json.loads(self.request.body)
                setattr('__cached_body', json_body)
                return json_body
            except Exception as err:
                logging.warning('Request is missing a body: %s -> %s',
                                err.__class__.__name__, str(err))
        raise MissingBody('Request is missing a body')

    def get_route_args(self):
        """
        Parses the route arguments based on the router generated url

        :return: route arguments in a :py:class:`dict` instance.

        For instance if the view url is ``r'/resource/(?P<slug>[^/]+)/'`` and the calling url is ``GET /resource/my-slug/``

        The returned dict will be formatted like this

        .. code-block:: python

           {'slug': 'my-slug'}

        .. note::
            Those route arguments are then available in each of the Handler methods in the self.route_args attribute

        """
        return self.request.route.regex.search(self.request.upath_info).groupdict()

    @classmethod
    def get_extra_actions(cls):
        actions = []

        for _method in cls.__dict__.values():
            if getattr(_method, 'is_dynamic_action', False):
                actions.append(_method)
        return actions


class APIModelBaseHandler(BaseHandler):
    """

    """
    lookup_field = 'urlsafe'
    lookup_value_regex = '[^/.]+'
    queryset = None
    serializer_class = None
    pagination_class = load_class(DEFAULT_PAGINATION_CLASS)

    def check_object_permissions(self, request, obj):
        """
        Returns always True by default. Override this method to
        get a object level permissions checks
        """
        return True

    def get_object(self, urlsafe=None, query_param=None, lookup_field=None, raise_exception=True):
        _key = None
        _lookup_field = lookup_field or self.lookup_field
        _query_param = query_param or self.query_param

        if _lookup_field is None and urlsafe is None and _query_param is None:
            if raise_exception is True:
                raise BadRequest
            else:
                return None

        try:
            if urlsafe is None:
                if _lookup_field is not None and _lookup_field in self.route_args:
                    urlsafe_from_request = self.route_args[_lookup_field]
                elif _query_param is not None:
                    urlsafe_from_request = self.request.GET.get(_query_param, None)
                else:
                    urlsafe_from_request = None

                if urlsafe_from_request is None and raise_exception is False:
                    return None

                if urlsafe_from_request is None:
                    raise MissingParameter('Parameter {} is missing from request'.format(
                        _lookup_field or _query_param
                    ))
                else:
                    urlsafe = urlsafe_from_request

            _key = get_key_from_urlsafe(urlsafe)
            if not self.check_object_permissions(self.request, _key):
                raise NotAuthorized('You are not authorized to access this resource')
        except Exception as err:
            logging.error('Error in get_object: %s -> %s\n%s',
                          err.__class__.__name__,
                          str(err),
                          traceback.format_exc())
            if raise_exception:
                raise NotFound
        finally:
            return _key

    @property
    def paginator(self):
        """
        The paginator instance associated with the view, or `None`.
        """
        if not hasattr(self, '_paginator'):
            if self.pagination_class is None:
                self._paginator = None
            else:
                self._paginator = self.pagination_class()
        return self._paginator

    def paginate_queryset(self, queryset):
        """
        Return a single page of results, or `None` if pagination is disabled.
        """
        if self.paginator is None:
            return None
        return self.paginator.paginate_queryset(queryset, self.request, view=self)

    def get_paginated_response(self, data):
        """
        Return a paginated style `Response` object for the given output data.
        """
        assert self.paginator is not None
        return self.paginator.get_paginated_response(data)

    def get_queryset(self):
        """
        Method used to retrieve the queryset. Must be overloaded for custom operations
        """
        return self.queryset

    def get_serializer(self, *args, **kwargs):
        """
        Method used to retrieves the serializer. Must be overloaded for custom operations
        """
        return self.serializer_class

    def post_save(self, obj, created=True):
        """
        Method called after each call to .put()
        """
        pass

    def pre_delete(self, obj):
        """
        Method called just before calling key.delete()
        """
        pass


class APIHandler(BaseHandler):
    def options(self, *args, **kwargs):
        return CorsResponse()

    def get(self, *args, **kwargs):
        raise NotImplemented

    def post(self, *args, **kwargs):
        raise NotImplemented

    def patch(self, *args, **kwargs):
        raise NotImplemented

    def delete(self, *args, **kwargs):
        raise NotImplemented

    def head(self, *args, **kwargs):
        raise NotImplemented

    def trace(self, *args, **kwargs):
        raise NotImplemented


class APIModelListHandler(APIModelBaseHandler):
    """
    ModelHandler that perform a generic list operation
    """
    def list(self, request):
        """
        List operation

        :return: :py:class:`restae.response.JsonResponse`
        """
        page = self.paginate_queryset(self.get_queryset())

        if page is not None:
            serializer = self.get_serializer()(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer()(self.get_queryset(), many=True)
        return Response(serializer.data)


class APIModelCreateHandler(APIModelBaseHandler):
    """
    ModelHandler that perform a generic create operation
    """
    def create(self, request):
        """

        :return: :py:class:`restae.response.JsonResponse`
        """
        _model = get_model_class_from_query(self.get_queryset())
        obj = _model(**self.get_serializer()(data=self.get_body()).data)
        obj.put()
        self.post_save(obj, created=True)
        return JsonResponse(data=self.get_serializer()(obj).data)


class APIModelRetrieveHandler(APIModelBaseHandler):
    """
    ModelHandler that perform a generic get operation
    """
    def retrieve(self, request, key=None):
        """

        :return: :py:class:`restae.response.JsonResponse`
        """
        try:
            obj = key.get()
            if obj is None:
                raise NotFound
        except Exception:
            raise NotFound

        return JsonResponse(data=self.get_serializer()(obj).data)


class APIModelUpdateHandler(APIModelBaseHandler):
    """
    ModelHandler that perform a generic update operation
    """
    def update(self, request, key=None):
        try:
            obj = key.get()
            obj.update(self.get_body())
            obj.put()
            self.post_save(obj, created=False)
            return JsonResponse(data=self.get_serializer()(obj).data)
        except Exception as err:
            raise BadRequest(str(err))


class APIModelPatchHandler(APIModelBaseHandler):
    """
    ModelHandler that perform a generic patch operation
    """
    def partial_update(self, request, key=None):
        try:
            obj = key.get()
            obj.update(self.get_body())
            obj.put()
            self.post_save(obj, created=False)
            return JsonResponse(data=self.get_serializer()(obj).data)
        except Exception as err:
            raise BadRequest(str(err))


class APIModelDestroyHandler(APIModelBaseHandler):
    """
    ModelHandler that perform a generic delete operation
    """
    def destroy(self, request, key=None):
        """
        :return: :py:class:`restae.response.JsonResponse`
        """
        self.pre_delete(key)
        key.delete()
        return JsonResponse()


class APIModelMixinHandler(APIModelListHandler,
                           APIModelCreateHandler,
                           APIModelDestroyHandler,
                           APIModelRetrieveHandler,
                           APIModelUpdateHandler,
                           APIModelPatchHandler):
    pass


class APIModelHandler(APIModelMixinHandler):
    def do_dispatch(self):
        """
        Dispatches the request.

        This will first check if there's a handler_method defined in the
        matched route, and if not it'll use the method correspondent to the
        request method (``get()``, ``post()`` etc).
        """

        if self.request.method.upper() == 'OPTIONS':
            return CorsResponse()

        route_args = self.get_route_args()
        route_kwargs = {}
        matched_url = None

        for url in getattr(self, 'urls', []):
            if re.match(url.regex, self.request.path_info):
                matched_url = url
                break

        if matched_url is None:
            raise NotFound('URL not found on this server')

        if matched_url.route.detail is True:
            try:
                route_kwargs['key'] = self.get_object(urlsafe=route_args['urlsafe'])
            except Exception:
                return JsonResponse(status=400, data='Given urlsafe is invalid')

        if self.request.method.lower() not in matched_url.route.mapping.keys():
            raise DispatchError('Invalid method {}'.format(self.request.method))

        self.action = matched_url.route.mapping[self.request.method.lower()]
        method = getattr(self, self.action, None)
        # The handler only receives *args if no named variables are set.
        args, kwargs = route_args, route_kwargs
        if kwargs:
            args = ()

        if self.check_action_permissions() is False:
            raise NotAuthorized()
        return method(self, *args, **kwargs)



"""
Middleware tests
"""
import unittest
import sys
import webapp2
import webtest

import dev_appserver
from webob import Response

sys.path = dev_appserver.EXTRA_PATHS + sys.path

from google.appengine.ext import testbed, ndb

from restae.handlers import APIHandler
from restae.middleware import Middleware
from restae.router import Router
from restae.conf import settings


class TestRequestMiddleware(Middleware):
    def process_request(self, request):
        request.value = 'request'


class TestResponseMiddleware(Middleware):
    def process_response(self, request, response):
        response.body = 'response'


class APIUserHandler(APIHandler):
    def get(self, *args, **kwargs):
        return Response(body=self.request.value)

    def post(self, *args, **kwargs):
        return Response()


class APIHandlerTestCase(unittest.TestCase):
    def setUp(self):
        self.testbed = testbed.Testbed()
        self.testbed.activate()
        self.testbed.init_datastore_v3_stub()
        self.testbed.init_memcache_stub()

        ndb.get_context().clear_cache()

        router = Router()
        router.register('user', APIUserHandler)
        app = webapp2.WSGIApplication(router.urls)
        self.testapp = webtest.TestApp(app)

    def test_request_middleware(self):
        settings.override({
            'MIDDLEWARE_CLASSES': ['restae.middleware_test.TestRequestMiddleware']
        })

        response = self.testapp.get('/user/')
        self.assertEquals(response.body, 'request')

    def test_response_middleware(self):
        settings.override({
            'MIDDLEWARE_CLASSES': ['restae.middleware_test.TestResponseMiddleware']
        })

        response = self.testapp.post('/user/')
        self.assertEquals(response.body, 'response')


if __name__ == '__main__':
    unittest.main()



"""
Serializers test
"""
import unittest
import sys

import dev_appserver

from restae.decorators import action
from restae.response import JsonResponse

sys.path = dev_appserver.EXTRA_PATHS + sys.path


from restae.router import Router
from restae.handlers import APIHandler, APIModelHandler


class Handler(APIHandler):
    pass


class ModelHandler(APIModelHandler):
    pass


class DynamicModelHandler(APIModelHandler):
    @action(methods=['GET'], detail=False)
    def toto(self, *args, **kwargs):
        return JsonResponse(data={'ok': 42})


class BaseTestCase(unittest.TestCase):
    def test_dynamic_route(self):
        router = Router()
        handler_name = 'handler'
        router.register(handler_name, DynamicModelHandler)

        self.assertItemsEqual(
            [
                ('^/{}/$'.format(handler_name), DynamicModelHandler),
                ('^/{}$'.format(handler_name), DynamicModelHandler),
                ('^/{}/toto/$'.format(handler_name), DynamicModelHandler),
                ('^/{}/toto$'.format(handler_name), DynamicModelHandler),
                ('^/{}/(?P<urlsafe>[^/.]+)/$'.format(handler_name), DynamicModelHandler),
                ('^/{}/(?P<urlsafe>[^/.]+)$'.format(handler_name), DynamicModelHandler)
            ],
            router.urls
        )

    def test_basic_router(self):
        router = Router()
        handler_name = 'handler'
        router.register(handler_name, Handler)

        self.assertItemsEqual(
            [
                ('^/{}$'.format(handler_name), Handler),
                ('^/{}/$'.format(handler_name), Handler),
                ('^/{}/(?P<urlsafe>[^/.]+)$'.format(handler_name), Handler),
                ('^/{}/(?P<urlsafe>[^/.]+)/$'.format(handler_name), Handler)
            ],
            router.urls
        )

    def test_model_router(self):
        router = Router()
        handler_name = 'handler'
        router.register(handler_name, ModelHandler)

        self.assertItemsEqual(
            [
                ('^/{}/$'.format(handler_name), ModelHandler),
                ('^/{}$'.format(handler_name), ModelHandler),
                ('^/{}/(?P<urlsafe>[^/.]+)/$'.format(handler_name), ModelHandler),
                ('^/{}/(?P<urlsafe>[^/.]+)$'.format(handler_name), ModelHandler)
            ],
            router.urls
        )


if __name__ == '__main__':
    unittest.main()



"""
Serializers
"""
import datetime
import collections

from dateutil import parser
from google.appengine.ext.ndb import ModelAttribute

from exceptions import ValidationError, SerializerError
from helpers import get_key_from_urlsafe, KEY_CLASS


def get_attribute(obj, name):
    if issubclass(obj.__class__, collections.Mapping):
        if name in obj:
            return obj[name]
    attr = getattr(obj, name, None)

    if attr is None:
        raise SerializerError('Missing attribute {} on {}'.format(
            name, obj
        ))
    return attr


def has_attribute(obj, name):
    if issubclass(obj.__class__, collections.Mapping):
        if name in obj:
            return True
    attr = getattr(obj, name, None)

    if attr is None:
        return False
    return True


class Field(object):
    base_type = None

    def __init__(self, *args, **kwargs):
        self.is_many = kwargs.get('many', False)
        self.object = None if len(args) == 0 else args[0]
        self._data = kwargs.get('data', None)
        self.args = args
        self.kwargs = kwargs

        self.required = kwargs.get('required', False)
        self.source = kwargs.get('source', None)

    def input(self, value):
        if self.base_type is not None:
            try:
                return self.base_type(value)
            except Exception:
                raise ValidationError('Field: Unknown {} value: {}'.format(
                    self.__class__.__name__, value))
        return value

    def input_many(self, values):
        if self.is_many:
            return [
                self.input(value)
                for value in values
            ]
        else:
            return self.input(values)

    def output(self, value):
        return value

    def output_many(self, values):
        if self.is_many:
            return [
                self.output(value)
                for value in values
            ]
        else:
            return self.output(values)

    def validate_base_type(self, value):
        if self.base_type is not None:
            return isinstance(value, self.base_type)
        return True

    def serialize(self):
        """
        ---> si y'a un data={} on veux verifier que les fields sont bien types, pas de manquant etc..
        ---> si pas de data, on a un objet et on veux le mettre en dict
        """
        if self._data is None and self.object is not None:
            return self.output_many(self.object)
        elif self.object is None and self._data is not None:
            return self.input_many(self._data)
        else:
            raise SerializerError('If no positional argument are passed, you must use data= kwarg')

    @property
    def data(self):
        if self.is_many:
            return [
                self.__class__(obj, data=self._data).serialize()
                for obj in self.object
            ]
        return self.serialize()


class BooleanField(Field):
    base_type = bool
    TRUE_VALUES = [
        't', 'T',
        'y', 'Y', 'yes', 'YES',
        'true', 'True', 'TRUE',
        'on', 'On', 'ON',
        '1', 1,
        True
    ]
    FALSE_VALUES = [
        'f', 'F',
        'n', 'N', 'no', 'NO',
        'false', 'False', 'FALSE',
        'off', 'Off', 'OFF',
        '0', 0, 0.0,
        False
    ]

    def input(self, value):
        if value in self.TRUE_VALUES:
            return True
        elif value in self.FALSE_VALUES:
            return False
        raise ValidationError('Field: Unknown Boolean value: {}'.format(value))


class IntegerField(Field):
    base_type = int


class StringField(Field):
    base_type = str


class FloatField(Field):
    base_type = float


class KeyField(Field):
    base_type = KEY_CLASS

    def input(self, value):
        try:
            return get_key_from_urlsafe(value)
        except Exception:
            raise ValidationError('Field: Unknown Key value: {}'.format(value))

    def output(self, value):
        if not self.validate_base_type(value) and value is not None:
            raise ValidationError('Field: Needs a Key class, {} given'.format(value.__class__.__name__))
        return value.urlsafe() if value else None


class DatetimeField(Field):
    base_type = datetime.datetime

    def input(self, value):
        try:
            if 'format' in self.kwargs:
                return datetime.datetime.strptime(value, self.kwargs['format'])
            return parser.parse(value)
        except Exception:
            raise ValidationError('Field: Unknown {} value: {}'.format(
                self.__class__.__name__, value))

    def output(self, value):
        if 'format' in self.kwargs:
            return value.strftime(value, self.kwargs['format'])
        return value.isoformat()


class MethodField(Field):
    pass


TYPE_MAPPING = {
    'IntegerProperty': IntegerField,
    'FloatProperty': FloatField,
    'BooleanProperty': BooleanField,
    'StringProperty': StringField,
    'TextProperty': StringField,
    'DateTimeProperty': DatetimeField,
    'KeyProperty': KeyField
}

READ_ONLY_FIELDS = [
    MethodField
]


class Serializer(Field):
    def get_class_fields(self, read_only=False):
        _attrs = {}

        for key, value in self.__class__.__dict__.iteritems():
            if issubclass(value.__class__, Field):
                if read_only is False:
                    if value.__class__ not in READ_ONLY_FIELDS:
                        _attrs[key] = value
                else:
                    _attrs[key] = value
        return _attrs

    def input(self, data):
        payload = {}

        for attr, field in self.get_class_fields().iteritems():
            _source = field.source or attr
            if not has_attribute(data, _source) and field.required:
                raise SerializerError('{} requires field {} but is missing'.format(
                    self.__class__.__name__,
                    attr
                ))

            if has_attribute(data, _source):
                payload.update({
                    attr: field.input_many(get_attribute(data, _source))
                })

        return payload

    def output(self, obj):
        payload = {}

        for attr, field in self.get_class_fields(read_only=True).iteritems():
            _source = field.source or attr

            if field.__class__ == MethodField:
                _method = getattr(self, 'get_{}'.format(attr), None)
                if _method is None:
                    raise SerializerError('missing method get_{} for field MethodField'.format(attr))
                payload.update({
                    attr: _method(obj)
                })
            else:
                payload.update({
                    attr: field.output_many(getattr(obj, _source))
                })
        payload['key'] = obj.key.urlsafe()
        return payload


class ModelSerializer(Serializer):
    def get_class_fields(self, read_only=False):
        _attrs = {}

        _meta = getattr(self, 'Meta', None)
        if _meta.fields == '__all__':
            _fields = []
            for key, val in dict(_meta.model.__dict__).iteritems():
                if issubclass(val.__class__, ModelAttribute):
                    _fields.append(key)
        else:
            _fields = _meta.fields

        for field in _fields:
            if field == 'key':
                _attrs[field] = KeyField()
            else:
                try:
                    if isinstance(_meta.model, tuple):
                        _meta.model = _meta.model[0]
                    if field.capitalize() in _meta.model._properties:
                        _attrs[field] = TYPE_MAPPING[_meta.model._properties[field.capitalize()].__class__.__name__]()
                    else:
                        _attrs[field] = TYPE_MAPPING[_meta.model._properties[field].__class__.__name__]()
                except KeyError:
                    if getattr(self, 'get_{}'.format(field), None) is not None:
                        _attrs[field] = MethodField()
                    else:
                        raise SerializerError('Field {} doest not exists on {} or its type is not supported yet'.format(
                            field, _meta.model.__name__
                        ))
        return _attrs




"""
Paginator working with datastore
"""

from collections import OrderedDict

from google.appengine.api.datastore_errors import BadValueError
from google.appengine.ext.ndb import Cursor

from restae.response import JsonResponse
from restae.conf import settings, DEFAULT_PAGE_SIZE
from restae.exceptions import InvalidPage, NotFound, BadRequest


class BasePagination(object):
    display_page_controls = False

    def paginate_queryset(self, queryset, request, view=None):
        raise NotImplementedError('paginate_queryset() must be implemented.')

    def get_paginated_response(self, data):  # pragma: no cover
        raise NotImplementedError('get_paginated_response() must be implemented.')

    def get_results(self, data):
        return data['results']

    def get_page_size(self, request):
        return request.GET.get('page_size', settings.get('PAGE_SIZE', DEFAULT_PAGE_SIZE))


class CursorPagination(BasePagination):
    """
    A simple cursor based style that supports cursors urlsafe as
    query parameters. For example:
    http://api.example.org/accounts/?page_token=<URLSAFE STRING>
    """
    def __init__(self):
        self.count = 0
        self.page = None
        self.next_page = None
        self.has_next = None
        self.request = None

    def get_page_token(self):
        return self.request.GET.get('page_token', None)

    def paginate_queryset(self, queryset, request, view=None):
        """
        Paginate a queryset if required, either returning a
        page object, or `None` if pagination is not configured for this view.
        """
        self.count = queryset.count()
        self.request = request
        try:
            self.page, self.next_page, self.has_next = queryset.fetch_page(
                self.get_page_size(request), start_cursor=Cursor(urlsafe=self.get_page_token()))
        except InvalidPage:
            raise NotFound('Requested page not found')
        except BadValueError as err:
            raise BadRequest(str(err))

        return list(self.page)

    def get_paginated_response(self, data):
        return JsonResponse(data=OrderedDict([
            ('count', self.count),
            ('next', self.get_next_link()),
            # ('previous', self.get_previous_link()),
            ('results', data)
        ]))

    def get_next_link(self):
        if not self.has_next:
            return None
        return self.next_page.urlsafe()

    # def get_previous_link(self):
    #     if self.next_page is not None and self.get_page_token() is not None:
    #         return self.next_page.reversed().urlsafe()
    #     return None



"""
Handlers test
"""
import unittest
import sys
import webapp2
import webtest

import dev_appserver
from webob import Response
from webtest import AppError

from restae.decorators import action
from restae.response import JsonResponse

sys.path = dev_appserver.EXTRA_PATHS + sys.path

from google.appengine.ext import testbed, ndb
from mock import patch

from restae import serializers
from restae.handlers import APIModelHandler, APIHandler
from restae.router import Router


class UserModel(ndb.Model):
    email = ndb.StringProperty()
    last_name = ndb.StringProperty()
    first_name = ndb.StringProperty()
    age = ndb.IntegerProperty()


class UserModelSerializer(serializers.ModelSerializer):
    class Meta:
        fields = '__all__'
        model = UserModel


class UserHandler(APIModelHandler):
    queryset = UserModel.query()
    serializer_class = UserModelSerializer


class DynamicUserHandler(APIModelHandler):
    queryset = UserModel.query()
    serializer_class = UserModelSerializer

    @action(methods=['GET'], detail=False)
    def toto(self, *args, **kwargs):
        return Response(body='toto')

    @action(methods=['GET'], detail=True)
    def tata(self, request, key):
        return Response(body=key.get().email)


class DynamicAPIModelHandlerTestCase(unittest.TestCase):
    def setUp(self):
        self.testbed = testbed.Testbed()
        self.testbed.activate()
        self.testbed.init_datastore_v3_stub()
        self.testbed.init_memcache_stub()

        ndb.get_context().clear_cache()

        router = Router()
        router.register('user', DynamicUserHandler)
        app = webapp2.WSGIApplication(router.urls)
        self.testapp = webtest.TestApp(app)
        self.entity = UserModel(
            email='admin@restae.com',
            first_name='admin',
            last_name='restae',
            age=28
        )
        self.entity.put()

    def test_dispatch_toto(self):
        response = self.testapp.get('/user/toto')
        self.assertEquals(response.body, 'toto')

    def test_dispatch_toto_bad_method(self):
        response = self.testapp.put('/user/toto', expect_errors=True)
        self.assertEquals(response.status_code, 404)

    def test_dispatch_toto_slash(self):
        response = self.testapp.get('/user/toto/')
        self.assertEquals(response.body, 'toto')

    def test_dispatch_tata(self):
        response = self.testapp.get('/user/{}/tata'.format(self.entity.key.urlsafe()))
        self.assertEquals(response.body, 'admin@restae.com')

    def test_dispatch_tata_with_slash(self):
        response = self.testapp.get('/user/{}/tata/'.format(self.entity.key.urlsafe()))
        self.assertEquals(response.body, 'admin@restae.com')

    def test_dispatch_tata_bad_method(self):
        response = self.testapp.post('/user/{}/tata'.format(self.entity.key.urlsafe()), expect_errors=True)
        self.assertEquals(response.status_code, 404)


class APIModelHandlerTestCase(unittest.TestCase):
    def setUp(self):
        self.testbed = testbed.Testbed()
        self.testbed.activate()
        self.testbed.init_datastore_v3_stub()
        self.testbed.init_memcache_stub()

        ndb.get_context().clear_cache()

        router = Router()
        router.register('user', UserHandler)
        app = webapp2.WSGIApplication(router.urls)
        self.testapp = webtest.TestApp(app)
        self.entity = UserModel(
            email='admin@restae.com',
            first_name='admin',
            last_name='restae',
            age=28
        )
        self.entity.put()

    @patch('restae.handlers.APIModelHandler.list', return_value=Response())
    def test_dispatch_list(self, mock_list):
        self.testapp.get('/user/')
        mock_list.assert_called_once()

    @patch('restae.handlers.APIModelHandler.create', return_value=Response())
    def test_dispatch_create(self, mock_create):
        self.testapp.post_json('/user/', {})
        mock_create.assert_called_once()

    @patch('restae.handlers.APIModelHandler.partial_update', return_value=Response())
    def test_dispatch_partial_update(self, mock_update):
        with self.assertRaises(AppError):
            response = self.testapp.put_json('/user/', {})
            self.assertEquals(response.status_code, 404)

        self.testapp.patch_json('/user/{}/'.format(self.entity.key.urlsafe()), {})
        mock_update.assert_called_once()

    @patch('restae.handlers.APIModelHandler.update', return_value=Response())
    def test_dispatch_update(self, mock_update):
        self.testapp.put_json('/user/{}/'.format(self.entity.key.urlsafe()), {})
        mock_update.assert_called_once()

    @patch('restae.handlers.APIModelHandler.destroy', return_value=Response())
    def test_dispatch_destroy(self, mock_destroy):
        self.testapp.delete('/user/{}/'.format(self.entity.key.urlsafe()))
        mock_destroy.assert_called_once()

    @patch('restae.handlers.APIModelHandler.retrieve', return_value=Response())
    def test_dispatch_retrieve(self, mock_retrieve):
        self.testapp.get('/user/{}/'.format(self.entity.key.urlsafe()))
        mock_retrieve.assert_called_once()


class APIUserHandler(APIHandler):
    def get(self, *args, **kwargs):
        return Response(body='get')

    def post(self, *args, **kwargs):
        return Response(body='post')

    def put(self, *args, **kwargs):
        return Response(body='put')

    def delete(self, *args, **kwargs):
        return Response(body='delete')

    def patch(self, *args, **kwargs):
        return Response(body='patch')


class APIHandlerTestCase(unittest.TestCase):
    def setUp(self):
        self.testbed = testbed.Testbed()
        self.testbed.activate()
        self.testbed.init_datastore_v3_stub()
        self.testbed.init_memcache_stub()

        ndb.get_context().clear_cache()

        router = Router()
        router.register('user', APIUserHandler)
        app = webapp2.WSGIApplication(router.urls)
        self.testapp = webtest.TestApp(app)

    def test_dispatch_get(self):
        response = self.testapp.get('/user/')
        self.assertEquals(response.body, 'get')

    def test_dispatch_post(self):
        response = self.testapp.post('/user/')
        self.assertEquals(response.body, 'post')

    def test_dispatch_put(self):
        response = self.testapp.put('/user/')
        self.assertEquals(response.body, 'put')

    def test_dispatch_patch(self):
        response = self.testapp.patch('/user/')
        self.assertEquals(response.body, 'patch')

    def test_dispatch_delete(self):
        response = self.testapp.delete('/user/')
        self.assertEquals(response.body, 'delete')


if __name__ == '__main__':
    unittest.main()



"""
Router
"""
import itertools

from restae.exceptions import ImproperlyConfigured


def escape_curly_brackets(url_path):
    """
    Double brackets in regex of url_path for escape string formatting
    """
    if ('{' and '}') in url_path:
        url_path = url_path.replace('{', '{{').replace('}', '}}')
    return url_path


class Route(object):
    """
    A route mapping representation contains the URL, METHOD and route type
    """
    def __init__(self, **kwargs):
        self.url = kwargs.pop('url')
        self.mapping = kwargs.get('mapping')
        self.detail = kwargs.pop('detail')
        self.is_dynamic = kwargs.pop('is_dynamic')


class Url(object):
    """
    An URL representation contains the Route and the destination Handler
    """
    def __init__(self, **kwargs):
        self.route = kwargs.pop('route')
        self.regex = kwargs.pop('regex')
        self.handler_class = kwargs.pop('handler_class')
        self.handler_name = kwargs.pop('handler_name')

    def copy(self):
        return Url(
            route=self.route,
            regex=self.regex,
            handler_class=self.handler_class,
            handler_name=self.handler_name
        )


class Router(object):
    """
    The default router
    """
    routes = [
        # List route.
        Route(
            url=r'^/{prefix}',
            mapping={
                'get': 'list',
                'post': 'create'
            },
            detail=False,
            is_dynamic=False
        ),
        # Dynamically generated list routes. Generated using
        # @action(detail=False) decorator on methods of the handler.
        Route(
            url=r'^/{prefix}/{url_path}',
            detail=False,
            is_dynamic=True
        ),
        # Detail route.
        Route(
            url=r'^/{prefix}/{lookup}',
            mapping={
                'get': 'retrieve',
                'put': 'update',
                'patch': 'partial_update',
                'delete': 'destroy'
            },
            detail=True,
            is_dynamic=False
        ),
        # Dynamically generated detail routes. Generated using
        # @action(detail=True) decorator on methods of the handler.
        Route(
            url=r'^/{prefix}/{lookup}/{url_path}',
            detail=True,
            is_dynamic=True
        ),
    ]

    def __init__(self):
        self.registry = []

    def register(self, handler_name, handler_class):
        assert handler_name is not None and handler_class is not None, (
            '\'handler_name\' and \'handler_class\' arguments are required '
            'when calling \'register\' function from router.'
        )
        self.registry.append((handler_name, handler_class))

    def get_lookup_regex(self, handler):
        """
        Given a handler, return the portion of URL regex that is used
        to match against a single instance.
        """
        base_regex = '(?P<{lookup_url_kwarg}>{lookup_value})'
        # Use `urlsafe` as default field.
        lookup_field = getattr(handler, 'lookup_field', None) or 'urlsafe'
        lookup_url_kwarg = getattr(handler, 'lookup_url_kwarg', None) or lookup_field
        lookup_value = getattr(handler, 'lookup_value_regex', '[^/.]+')
        return base_regex.format(
            lookup_url_kwarg=lookup_url_kwarg,
            lookup_value=lookup_value
        )

    def _get_dynamic_route(self, route, action):
        url_path = escape_curly_brackets(action.url_path)

        return Route(
            url=route.url.replace('{url_path}', url_path),
            mapping=action.mapping,
            detail=route.detail,
            is_dynamic=True
        )

    def get_routes(self, handler):
        """
        Augment `self.routes` with any dynamically generated routes.
        Returns a list of the Route class.
        """
        # converting to list as iterables are good for one pass, known host needs to be checked again and again for
        # different functions.
        known_actions = list(itertools.chain(*[route.mapping.values() for route in self.routes if not route.is_dynamic]))
        extra_actions = handler.get_extra_actions()

        # checking action names against the known actions list
        not_allowed = [
            action.__name__ for action in extra_actions
            if action.__name__ in known_actions
        ]
        if not_allowed:
            msg = ('Cannot use the @action decorator on the following '
                   'methods, as they are existing routes: %s')
            raise ImproperlyConfigured(msg % ', '.join(not_allowed))

        # partition detail and list actions
        detail_actions = [action for action in extra_actions if action.detail]
        list_actions = [action for action in extra_actions if not action.detail]

        routes = []
        dynamic_routes = []
        for route in self.routes:
            if route.is_dynamic and route.detail:
                dynamic_routes += [self._get_dynamic_route(route, action) for action in detail_actions]
            elif route.is_dynamic and not route.detail:
                dynamic_routes += [self._get_dynamic_route(route, action) for action in list_actions]
            else:
                routes.append(route)

        return dynamic_routes + routes

    def get_method_map(self, handler, method_map):
        """
        Given an handler, and a mapping of http methods to actions,
        return a new mapping which only includes any mappings that
        are actually implemented by the viewset.
        """
        bound_methods = {}
        for method, action in method_map.items():
            if hasattr(handler, action):
                bound_methods[method] = action
        return bound_methods

    def get_urls(self):
        """
        Use the registered handlers to generate a list of URL patterns.
        """
        urls = []

        for handler_name, handler_class in self.registry:
            lookup = self.get_lookup_regex(handler_class)
            routes = self.get_routes(handler_class)
            ret = []

            for route in routes:

                # Only actions which actually exist on the viewset will be bound
                # mapping = self.get_method_map(handler_class, route.mapping)
                # if not mapping:
                #     continue

                # Build the url pattern
                regex = route.url.format(
                    prefix=handler_name,
                    lookup=lookup,
                )

                _url_with_slash = Url(
                    handler_class=handler_class,
                    handler_name=handler_name,
                    route=route,
                    regex=regex + '/$'
                )
                _url_without_slash = Url(
                    handler_class=handler_class,
                    handler_name=handler_name,
                    route=route,
                    regex=regex + '$'
                )

                ret.append(_url_with_slash)
                ret.append(_url_without_slash)
                urls.append(_url_with_slash)
                urls.append(_url_without_slash)

            setattr(handler_class, 'urls', ret)
        return urls

    @property
    def urls(self):
        if not hasattr(self, '_urls'):
            self._urls = []

            for _url in self.get_urls():
                self._urls.append((
                    _url.regex,
                    _url.handler_class
                ))
        return self._urls



import webapp2

from google.appengine.ext import ndb

from restae.handlers import APIModelHandler
from restae.router import Router
from restae.serializers import ModelSerializer
from restae.response import JsonResponse
from restae.decorators import action


class User(ndb.Model):
    email = ndb.StringProperty()
    first_name = ndb.StringProperty()
    last_name = ndb.StringProperty()


class UserModelSerializer(ModelSerializer):
    class Meta:
        model = User
        fields = '__all__'


class Handler(APIModelHandler):
    queryset = User.query()
    serializer_class = UserModelSerializer

    @action(methods=['GET'], detail=True)
    def toto(self, *args, **kwargs):
        return JsonResponse(data={'ok': 42})


router = Router()
router.register('user', Handler)
print router.urls
app = webapp2.WSGIApplication(router.urls)



"""
Base middleware class
"""
import webapp2


class Middleware(object):
    """
    Base middleware class
    """
    ACTIVATE_ON_METHODS = webapp2.WSGIApplication.allowed_methods.union(('PATCH',))

    def activate_on_method(self, method):
        return method.upper() in self.ACTIVATE_ON_METHODS

    def process_request(self, request):
        """
        Will be called before handler
        """
        pass

    def process_response(self, request, response):
        """
        Will be called after handler
        """
        pass



import webapp2

# Fix to allow the use of PATCH command with webapp2 base framework
webapp2.WSGIApplication.allowed_methods = webapp2.WSGIApplication.allowed_methods.union(('PATCH',))

__version__ = '1.1.5'



"""
Custom responses
"""
import six
import collections

import webob
import json

from restae.conf import settings


class CorsResponse(webob.Response):
    def __init__(self, *args, **kwargs):
        super(CorsResponse, self).__init__(*args, **kwargs)
        self.headers['Access-Control-Allow-Origin'] = ', '.join(settings.get('CORS_ALLOW_ORIGIN'))
        self.headers['Access-Control-Allow-Headers'] = ', '.join(settings.get('CORS_ALLOW_HEADERS'))
        self.headers['Access-Control-Allow-Methods'] = ', '.join(settings.get('CORS_ALLOW_METHODS'))


class JsonResponse(CorsResponse):
    def __init__(self, *args, **kwargs):
        data = None

        if 'data' in kwargs:
            data = kwargs.pop('data')
        super(JsonResponse, self).__init__(*args, **kwargs)
        self.headers['Content-type'] = 'application/json'

        if isinstance(data, (collections.Mapping, list, set, tuple)):
            self.body = json.dumps(data)
        elif isinstance(data, six.string_types):
            self.body = json.dumps({
                'message': data
            })
        elif data is not None:
            self.body = data



"""
Rest helpers
"""
import importlib
import threading

from google.appengine.ext import ndb

from restae.conf import settings

KEY_CLASS = ndb.Key


class cached_property(object):
    """A decorator that converts a function into a lazy property.

    The function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value::

        class Foo(object):

            @cached_property
            def foo(self):
                # calculate something important here
                return 42

    The class has to have a `__dict__` in order for this property to
    work.

    .. note:: Implementation detail: this property is implemented as non-data
       descriptor.  non-data descriptors are only invoked if there is
       no entry with the same name in the instance's __dict__.
       this allows us to completely get rid of the access function call
       overhead.  If one choses to invoke __get__ by hand the property
       will still work as expected because the lookup logic is replicated
       in __get__ for manual invocation.

    This class was ported from `Werkzeug`_ and `Flask`_.
    """

    _default_value = object()

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func
        self.lock = threading.RLock()

    def __get__(self, obj, type=None):
        if obj is None:
            return self

        with self.lock:
            value = obj.__dict__.get(self.__name__, self._default_value)
            if value is self._default_value:
                value = self.func(obj)
                obj.__dict__[self.__name__] = value

            return value


def get_key_from_urlsafe(urlsafe):
    """
    Build a Key with given urlsafe
    """
    return ndb.Key(urlsafe=urlsafe)


def get_model_class_from_query(query):
    """
    Return model class from any query.
    Note that the model needs to be imported once
    in the application
    """
    return ndb.Model._lookup_model(query.kind)


def get_object_from_urlsafe(urlsafe):
    """
    Build a Key with given urlsafe and get the object
    """
    return get_key_from_urlsafe(urlsafe)


def load_class(full_class_string):
    """
    Dynamically load a class from a string
    """
    class_data = full_class_string.split('.')
    module_path = '.'.join(class_data[:-1])
    class_str = class_data[-1]

    module = importlib.import_module(module_path)
    # Finally, we retrieve the Class
    return getattr(module, class_str, None)


def get_middlewares():
    """
    Return all middleware classes
    """
    middlewares = []

    for middleware in settings.get('MIDDLEWARE_CLASSES', []):
        middlewares.append(load_class(middleware)())

    return middlewares



"""
CLI commands
"""
import sys
import os


API = ('"""\n'
       'generated by restae-admin\n'
       '\n'
       '{module_name_cap} module\n'
       '"""\n'
       '\n'
       'from restae.handlers import APIModelHandler\n'
       'from model import {module_name_cap}, {module_name_cap}Serializer\n'
       '\n'
       '\n'
       'class {module_name_cap}Handler(APIModelHandler):\n'
       '    """\n'
       '    {module_name_cap} Handler\n'
       '    """\n'
       '    queryset = {module_name_cap}.query()\n'
       '    serializer_class = {module_name_cap}Serializer\n'
       '\n')

MODEL = ('"""\n'
         'generated by restae-admin\n'
         '\n'
         '{module_name_cap} models and serializers\n'
         '"""\n'
         '\n'
         'from google.appengine.ext import ndb\n'
         'from restae.serializers import ModelSerializer\n'
         '\n'
         '\n'
         'class {module_name_cap}(ndb.Model):\n'
         '    """\n'
         '    {module_name_cap} models class\n'
         '    """\n'
         '    pass\n'
         '\n'
         '\n'
         'class {module_name_cap}Serializer(ModelSerializer):\n'
         '    """\n'
         '    {module_name_cap} serializer\n'
         '    """\n'
         '    class Meta:\n'
         '        fields = \'__all__\'\n'
         '\n')


def generate(args):
    if args[0] == 'module':
        module_name = args[1]

        try:
            os.mkdir(module_name)
        except OSError:
            pass

        with open(os.path.join(module_name, '__init__.py'), 'w+') as ifile:
            ifile.write('# generated by restae-admin\n')

        with open(os.path.join(module_name, 'api.py'), 'w+') as ifile:
            ifile.write(API.format(
                module_name=module_name,
                module_name_cap=module_name.capitalize()
            ))

        with open(os.path.join(module_name, 'model.py'), 'w+') as ifile:
            ifile.write(MODEL.format(
                module_name=module_name,
                module_name_cap=module_name.capitalize()
            ))
        print 'Module ' + module_name + ' created !'


def execute_from_command_line(argv=None):
    args = argv or sys.argv
    args = args[1:]

    if args and len(args) == 3 and args[0] == 'g':
        generate(args[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

import restae

setup(
    name='restae',
    version=restae.__version__,
    packages=find_packages(),
    author="Camille TOLSA",
    author_email="camille.tolsa@gmail.com",
    description="REST Framwork for webapp2 & datastore applications",
    long_description=open('README.md').read(),
    url='http://github.com/tolsac/restae',
    include_package_data=True,
    install_requires=[],
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    zip_safe=False,
    licence='BSD',
    entry_points={'console_scripts': [
        'restae-admin = restae.management:execute_from_command_line',
    ]},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
    ]

)


"""
Serializers test
"""
import unittest
import sys
import datetime

import dev_appserver
sys.path = dev_appserver.EXTRA_PATHS + sys.path

from google.appengine.ext import testbed, ndb

from restae import serializers


class UserModel(ndb.Model):
    email = ndb.StringProperty()
    last_name = ndb.StringProperty()
    first_name = ndb.StringProperty()
    age = ndb.IntegerProperty()


class UserSerializer(serializers.Serializer):
    email = serializers.StringField()
    first_name = serializers.StringField()
    last_name = serializers.StringField()
    age = serializers.IntegerField()


class UserModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserModel
        fields = '__all__'


class AllFields(ndb.Model):
    float_val = ndb.FloatProperty()
    int_val = ndb.IntegerProperty()
    datetime_val = ndb.DateTimeProperty()
    bool_val = ndb.BooleanProperty()
    string_val = ndb.StringProperty()
    text_val = ndb.TextProperty()
    user_val = ndb.KeyProperty()


class AllFieldsSerializer(serializers.Serializer):
    float_val = serializers.FloatField()
    int_val = serializers.IntegerField()
    datetime_val = serializers.DatetimeField()
    bool_val = serializers.BooleanField()
    string_val = serializers.StringField()
    text_val = serializers.StringField()
    user_val = serializers.KeyField()


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        self.testbed = testbed.Testbed()
        self.testbed.activate()
        self.testbed.init_datastore_v3_stub()
        self.testbed.init_memcache_stub()

        ndb.get_context().clear_cache()


class SerializerTestCase(BaseTestCase):
    def test_serializer_output(self):
        entity = UserModel(
            email='admin@restae.com',
            first_name='admin',
            last_name='restae',
            age=28
        )
        entity.put()
        serialized = UserSerializer(entity).data
        self.assertDictEqual(
            serialized,
            {
                'key': entity.key.urlsafe(),
                'email': u'admin@restae.com',
                'first_name': u'admin',
                'last_name': u'restae',
                'age': 28
            }
        )

    def test_serializer_input(self):
        api_data = {
            'email': 'admin@restae.com',
            'first_name': 'admin',
            'last_name': 'restae',
            'age': 28
        }

        serialized = UserSerializer(data=api_data).data
        self.assertDictEqual(
            serialized,
            {
                'email': u'admin@restae.com',
                'first_name': u'admin',
                'last_name': u'restae',
                'age': 28
            }
        )

    def test_all_fields_output(self):
        user = UserModel(email='admin@restae.com')
        user.put()

        all_fields = AllFields(
            float_val=3.14,
            int_val=42,
            datetime_val=datetime.datetime(2018, 1, 1),
            bool_val=True,
            string_val='restae',
            text_val='framework',
            user_val=user.key
        )
        all_fields.put()

        self.assertDictEqual(
            AllFieldsSerializer(all_fields).data,
            {
                'key': all_fields.key.urlsafe(),
                'float_val': 3.14,
                'int_val': 42,
                'datetime_val': '2018-01-01T00:00:00',
                'bool_val': True,
                'string_val': u'restae',
                'text_val': u'framework',
                'user_val': user.key.urlsafe()
            }
        )

    def test_all_fields_input(self):
        user = UserModel(email='admin@restae.com')
        user.put()

        api_data = {
            'float_val': 3.14,
            'int_val': 42,
            'datetime_val': '2018-01-01T00:00:00',
            'bool_val': True,
            'string_val': 'restae',
            'text_val': 'framework',
            'user_val': user.key.urlsafe()
        }

        serialized = AllFieldsSerializer(data=api_data).data
        key_obj = serialized.pop('user_val')

        self.assertEqual(
            key_obj.__class__,
            ndb.Key
        )
        self.assertEqual(
            key_obj.kind(),
            'UserModel'
        )

        self.assertDictEqual(
            serialized,
            {
                'float_val': 3.14,
                'int_val': 42,
                'datetime_val': datetime.datetime(2018, 1, 1, 0, 0),
                'bool_val': True,
                'string_val': 'restae',
                'text_val': 'framework',
            }
        )

    def test_source_attr(self):
        user = UserModel(email='admin@restae.com')
        user.put()

        class UserSerializer(serializers.Serializer):
            contact = serializers.StringField(source='email')

        serialized = UserSerializer(user).data

        self.assertDictEqual(
            serialized,
            {
                'key': user.key.urlsafe(),
                'contact': 'admin@restae.com'
            }
        )

    def test_serializer_method(self):
        user = UserModel(email='admin@restae.com')
        user.put()

        class UserMethodSerializer(serializers.Serializer):
            toto = serializers.MethodField()

            def get_toto(self, obj):
                return 42

        serialized = UserMethodSerializer(user).data
        self.assertDictEqual(
            serialized,
            {
                'key': user.key.urlsafe(),
                'toto': 42
            }
        )

    def test_serializer_method_2(self):
        user = UserModel(email='admin@restae.com')
        user.put()

        class UserMethodSerializer(serializers.Serializer):
            contact = serializers.MethodField()

            def get_contact(self, obj):
                return obj.email

        serialized = UserMethodSerializer(user).data
        self.assertDictEqual(
            serialized,
            {
                'key': user.key.urlsafe(),
                'contact': 'admin@restae.com'
            }
        )


class ModelSerializerTestCase(BaseTestCase):
    def test_model_serializer_output(self):
        entity = UserModel(
            email='admin@restae.com',
            first_name='admin',
            last_name='restae',
            age=28
        )
        entity.put()
        serialized = UserModelSerializer(entity).data

        self.assertDictEqual(
            serialized,
            {
                'key': entity.key.urlsafe(),
                'email': u'admin@restae.com',
                'first_name': u'admin',
                'last_name': u'restae',
                'age': 28
            }
        )

    def test_model_serializer_with_filter_output(self):
        entity = UserModel(
            email='admin@restae.com',
            first_name='admin',
            last_name='restae',
            age=28
        )
        entity.put()
        UserModelSerializer.Meta.fields = ('key', 'email')
        serialized = UserModelSerializer(entity).data

        self.assertDictEqual(
            serialized,
            {
                'key': entity.key.urlsafe(),
                'email': u'admin@restae.com'
            }
        )

    def test_with_custom_attrs(self):
        class UserDetailSerializer(UserModelSerializer):
            phones = serializers.MethodField()

            def get_phones(self, obj):
                return ['0622380111']

            class Meta:
                model = UserModel,
                fields = ['first_name', 'last_name', 'email', 'age', 'phones']

        entity = UserModel(
            email='admin@restae.com',
            first_name='admin',
            last_name='restae',
            age=28
        )
        entity.put()
        serialized = UserDetailSerializer(entity).data

        self.assertDictEqual(
            serialized,
            {
                'key': entity.key.urlsafe(),
                'email': u'admin@restae.com',
                'first_name': u'admin',
                'last_name': u'restae',
                'age': 28,
                'phones': ['0622380111']
            }
        )

    def test_key_in_key(self):
        class AllFieldsSerializer(serializers.ModelSerializer):
            class Meta:
                model = AllFields
                fields = '__all__'

        user = UserModel(email='admin@restae.com')
        user.put()

        all_fields = AllFields(**{
            'float_val': 3.14,
            'int_val': 42,
            'datetime_val': datetime.datetime(2018, 1, 1, 0, 0),
            'bool_val': True,
            'string_val': 'restae',
            'text_val': 'framework',
            'user_val': user.key
        })
        all_fields.put()

        serialized = AllFieldsSerializer(all_fields).data
        self.assertDictEqual(
            serialized,
            {
                'float_val': 3.14,
                'int_val': 42,
                'datetime_val': '2018-01-01T00:00:00',
                'bool_val': True,
                'string_val': u'restae',
                'text_val': u'framework',
                'user_val': user.key.urlsafe(),
                'key': all_fields.key.urlsafe()
            }
        )


if __name__ == '__main__':
    unittest.main()
